/***************************************************************************
 BioSequence.cc  -  description
 -------------------
 begin                : Thu Jun 16 2002
 copyright            : (C) 2002 by Oleksiy Kohany
 email                : kohany@girinst.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "BioSequence.h"
#include "baddata_exception.h"
#include <string.h>
#include <sstream>
using namespace std;

//#define DEBUG_ALLOC
const char FRAGMENT[] = "FRAGMENT";
/** ---------------------------------------------------------- */
BioSequence::BioSequence() {
   seqtype = SEQTYPE_UNKNOWN;
#ifdef DEBUG_ALLOC
   cerr << "In BioSequence(): " << endl;
#endif
}
/** ---------------------------------------------------------- */
BioSequence::BioSequence(const BioSequence& other) :
   string(other), m_sName(other.getName()), seqtype(other.seqtype) {
#ifdef DEBUG_ALLOC
   cerr << "In BioSequence(const BioSequence& other): " << other.getName() << " : " << other.size() << "bp" << endl;
#endif
}
/** ---------------------------------------------------------- */
BioSequence::BioSequence(const string &sSeq, const string &sName) :
   string(sSeq) {
   m_sName = sName;
   seqtype = SEQTYPE_UNKNOWN;
#ifdef DEBUG_ALLOC
   cerr << "In BioSequence(const string &sSeq, const string &sName): " << endl;
#endif
}
/** ---------------------------------------------------------- */
BioSequence::BioSequence(const char *pcSeq, const char *pcName) :
   string(pcSeq) {
   m_sName = pcName;
   seqtype = SEQTYPE_UNKNOWN;
#ifdef DEBUG_ALLOC
   cerr << "In BioSequence(const char *pcSeq, const char *pcName): " << endl;
#endif
}
/** ---------------------------------------------------------- */
BioSequence::~BioSequence() {
#ifdef DEBUG_ALLOC
   cerr << "In ~BioSequence() for " << m_sName << endl;
#endif
}

/** ====================================== */
void BioSequence::formatedPrint(ostream &os, int iLineLen) const {
   unsigned long iPos;
   for (iPos = 0; iPos < string::size(); iPos += iLineLen) {
      os << string::substr(iPos, iLineLen) << "\n";
   }
}
/** ====================================== */
void BioSequence::formatedPrint(FILE *FOUT, int iLineLen) const {
   unsigned long iPos;
   for (iPos = 0; iPos < string::size(); iPos += iLineLen) {
      fprintf(FOUT, "%s\n", string::substr(iPos, iLineLen).c_str());
   }
}
/** ====================================== */
void BioSequence::unformatedPrint(ostream &os) const {
   os << c_str() << endl;
}
/** ====================================== */
void BioSequence::unformatedPrint(FILE *FOUT) const {
   fprintf(FOUT, "%s\n", c_str());
}
/** ====================================== */
void BioSequence::rmTrash() {
   string::iterator ptr = begin();
   while (ptr != end()) {
      if (!isgraph(*ptr)) {
         ptr = erase(ptr);
      } else {
         ptr++;
      }
   }
   ptr--;
   if (*ptr == '1') {
      erase(ptr);
   }
}
/** ====================================== */
void BioSequence::clear() {
   seqtype = SEQTYPE_UNKNOWN;
   setName("");
   assign("");
}

/** ====================================== */
int BioSequence::removeGaps() {
   int gap_pos = 0;
   int cnt = 0;
   BioSequence::iterator sptr = begin();
   while (sptr != end()) {
      if (isalpha(*sptr) || *sptr == '*') {
         sptr++;
      } else {
         sptr = erase(sptr);
      }
   }
   return cnt;
}
BioSequence& BioSequence::operator =(const string& s) {
   string::assign(s);
   seqtype = SEQTYPE_UNKNOWN;
   return *this;
}
BioSequence& BioSequence::operator =(const char* s) {
   string::assign(s);
   seqtype = SEQTYPE_UNKNOWN;
   return *this;
}
BioSequence& BioSequence::append(const BioSequence& s) {
   append((string) s);
   if (getName().size() == 0) {
      setName(s.getName());
   }
   return *this;
}
BioSequence& BioSequence::append(const string& s) {
   string::append(s);
   //seqtype = SEQTYPE_UNKNOWN;
   return *this;
}
BioSequence& BioSequence::append(const char* s) {
   string::append(s);
   //seqtype = SEQTYPE_UNKNOWN;
   return *this;
}
BioSequence& BioSequence::assign(const char* s) {
   string::assign(s);
   seqtype = SEQTYPE_UNKNOWN;
   return *this;
}
BioSequence& BioSequence::assign(const char* s, size_type n) {
   string::assign(s, n);
   seqtype = SEQTYPE_UNKNOWN;
   return *this;
}
BioSequence& BioSequence::assign(const char *pcSeq, const char *pcName) {
   string::assign(pcSeq);
   rmTrash();
   m_sName = pcName;
   seqtype = SEQTYPE_UNKNOWN;
   return *this;
}
BioSequence& BioSequence::assign(const string& s) {
   string::assign(s);
   seqtype = SEQTYPE_UNKNOWN;
   return *this;
}
BioSequence& BioSequence::assign(const string &sSeq, const string &sName) {
   string::assign(sSeq);
   rmTrash();
   m_sName = sName;
   seqtype = SEQTYPE_UNKNOWN;
   return *this;
}
BioSequence& BioSequence::assign(const BioSequence &Seq) {
   string::assign(Seq);
   rmTrash();
   setName(Seq.getName());
   seqtype = Seq.seqtype;
   return *this;
}
/** ====================================== */
const string& BioSequence::getName() const {
   return m_sName;
}
/** ====================================== */
void BioSequence::setName(const string &sName) {
   m_sName = sName;
}
/** ====================================== */
void BioSequence::setName(const char *pcName) {
   m_sName = pcName;
}
/** ====================================== */
string& BioSequence::complement(string& str) {
   string::iterator rptr;
   string::iterator dptr;
   for (rptr = str.end(), rptr--, dptr = str.begin(); rptr >= dptr; rptr--, dptr++) {
      char c = *dptr;
      *dptr = complement(*rptr);
      *rptr = complement(c);
   }
   return str;
}
/** ====================================== */
BioSequence& BioSequence::complement() {
   rmTrash();
   complement(*this);
   return *this;
}
/** ====================================== */
BioSequence& BioSequence::operator =(const BioSequence &Seq) {
   assign(Seq);
   return *this;
}
/** ====================================== */
long BioSequence::findBase(long iPos) const {
   return string::find_first_not_of(GAPSYMB, iPos);
}
/** ====================================== */
BioSequence& BioSequence::append(char c) {
   string::append(1, c);
   return *this;
}
/** ====================================== */
BioSequence& BioSequence::sqeezeOutGaps() {
   string::iterator sptr, delptr;
   for (sptr = string::begin(); sptr < string::end(); sptr++) {
      if (!isalpha(*sptr)) {
         delptr = sptr;
         sptr--;
         string::erase(delptr);
      }
   }
   return *this;
}
/** ====================================== */
BioSequence& BioSequence::format() {
   sqeezeOutGaps();
   toupper();
   return *this;
}
/** ====================================== */
void BioSequence::printHead(ostream& os) const {
#ifdef FASTA
   printFASTAHead(os);
#else

   printIGHead(os);
#endif

}
/** ====================================== */
void BioSequence::printHead(FILE* FOUT) const {
#ifdef FASTA
   printFASTAHead(FOUT);
#else

   printIGHead(FOUT);
#endif

}
/** ====================================== */
void BioSequence::printIGHead(ostream& os) const {
   os << ";\n" << m_sName << "\n";
}
/** ====================================== */
void BioSequence::printIGHead(FILE* FOUT) const {
   fprintf(FOUT, ";\n%s\n", m_sName.c_str());
}
/** ====================================== */
void BioSequence::printFASTAHead(ostream& os) const {
   os << ">" << m_sName << "\n";
}
/** ====================================== */
void BioSequence::printFASTAHead(FILE* FOUT) const {
   fprintf(FOUT, ">%s\n", m_sName.c_str());
}
/** ====================================== */
void BioSequence::print(ostream& os, bool format) const {
   printHead(os);
   format ? formatedPrint(os) : unformatedPrint(os);
}
/** ====================================== */
void BioSequence::print(FILE *FOUT, bool format) const {
   printHead(FOUT);
   format ? formatedPrint(FOUT) : unformatedPrint(FOUT);
}
/** ====================================== */
void BioSequence::printIG(ostream& os, bool format) const {
   printIGHead(os);
   format ? formatedPrint(os) : unformatedPrint(os);
}
/** ====================================== */
void BioSequence::printIG(FILE *FOUT, bool format) const {
   printIGHead(FOUT);
   format ? formatedPrint(FOUT) : unformatedPrint(FOUT);
}
/** ====================================== */
void BioSequence::printFASTA(ostream& os, bool format) const {
   printFASTAHead(os);
   format ? formatedPrint(os) : unformatedPrint(os);
}
/** ====================================== */
void BioSequence::printFASTA(FILE *FOUT, bool format) const {
   printFASTAHead(FOUT);
   format ? formatedPrint(FOUT) : unformatedPrint(FOUT);
}
/** ====================================== */
char BioSequence::complement(char c) {
   switch (c) {
   case 'a':
      return 't';
   case 'A':
      return 'T';
   case 'u':
   case 't':
      return 'a';
   case 'U':
   case 'T':
      return 'A';
   case 'c':
      return 'g';
   case 'C':
      return 'G';
   case 'g':
      return 'c';
   case 'G':
      return 'C';
   case 'm':
      return 'k';
   case 'M':
      return 'K'; //M = A or C
   case 'k':
      return 'm';
   case 'K':
      return 'M'; //K = G or T
   case 'r':
      return 'y';
   case 'R':
      return 'Y'; //R = A or G
   case 'y':
      return 'r';
   case 'Y':
      return 'R'; //Y = C or T
   case 'v':
      return 'b';
   case 'V':
      return 'B'; //V = A or C or G
   case 'b':
      return 'v';
   case 'B':
      return 'V'; //B = C or G or T
   case 'h':
      return 'd';
   case 'H':
      return 'D'; //H = A or C or T
   case 'd':
      return 'h';
   case 'D':
      return 'H'; //D = A or G or T
   default:
      return c;
   }
}
/** ====================================== */
vector<char> BioSequence::string2vector(const string& s) {
   vector<char> v;
   for (string::const_iterator sptr = s.begin(); sptr != s.end(); sptr++) {
      v.push_back(*sptr);
   }
   return v;
}
/** ====================================== */
string BioSequence::vector2string(const vector<char>& s) {
   string v;
   for (vector<char>::const_iterator sptr = s.begin(); sptr != s.end(); sptr++) {
      v.push_back(*sptr);
   }
   return v;
}
/** ====================================== */
BioSubSequence& BioSubSequence::complement() {
   BioSequence::complement();
   long iTmp;
   iTmp = getFrom();
   setFrom(getTo());
   setTo(iTmp);
   return *this;
}
/** ====================================== */
BioSequence& BioSequence::toupper() {
   string::iterator dptr;
   rmTrash();
   for (dptr = string::begin(); dptr != string::end(); dptr++) {
      *dptr = ::toupper(*dptr);
   }
   return *this;
}
/** ====================================== */
BioSequence& BioSequence::tolower() {
   string::iterator dptr;
   rmTrash();
   for (dptr = string::begin(); dptr != string::end(); dptr++) {
      *dptr = ::tolower(*dptr);
   }
   return *this;
}
/** ====================================== */
int BioSequence::read(istream& is) throw() {
   clear();
   string sLine;

   if (readHead(is)) {
      while (is.peek() != ';' && is.peek() != '>' && !(getline(is, sLine).fail())) {
         append(sLine);
      }
      rmTrash();
      guessType();
      return size();
   } else {
      return -1;
   }
}
/** ====================================== */
int BioSequence::read(FILE* is) throw() {
   clear();
   char sLine[MAX_LINELENGTH];
   char *ptr;
   char c;
   if (readHead(is)) {
      while ((c = ungetc(fgetc(is), is)) != ';' && c != '>' && fgets(sLine, MAX_LINELENGTH, is) != 0) {
         if ((ptr = strchr(sLine, '\n')) != 0) {
            *ptr = '\0';
         }
         append(sLine);
      }
      rmTrash();
      guessType();
      return size();
   } else {
      return -1;
   }
}
/** ======================================
 * gaplen - returns number of gaps skipped
 * */
string::const_iterator BioSequence::next_base(string::const_iterator ptr, int* gaplen) const {
   int gapnum = 0;
   while (++ptr != this->end()) {
      if (*ptr != GAPSYMB) {
         break;
      }
      gapnum++;
   }
   if (gaplen != NULL) {
      *gaplen = gapnum;
   }
   return ptr;
}
/** ======================================
 * gaplen - returns number of gaps skipped
 * */
string::const_iterator BioSequence::prev_base(string::const_iterator ptr, int* gaplen) const {
   int gapnum = 0;
   while (--ptr != this->begin()) {
      if (*ptr != GAPSYMB) {
         break;
      }
      gapnum++;
   }
   if (gaplen != NULL) {
      *gaplen = gapnum;
   }
   return ptr;
}
/** ======================================
 * gaplen - returns number of gaps skipped
 * */
char BioSequence::next_base(int p, int* gaplen) const {
   if (gaplen != NULL) {
      *gaplen = 0;
   }
   while (++p < this->size()) {
      if ((*this)[p] != GAPSYMB) {
         return (*this)[p];
      }
      if (gaplen != NULL) {
         (*gaplen)++;
      }
   }
   return 0;
}
/** ======================================
 * gaplen - returns number of gaps skipped
 * */
char BioSequence::prev_base(int p, int* gaplen) const {
   if (gaplen != NULL) {
      *gaplen = 0;
   }
   while (--p >= 0) {
      if ((*this)[p] != GAPSYMB) {
         return (*this)[p];
      }
      if (gaplen != NULL) {
         (*gaplen)++;
      }
   }
   return 0;
}
//==========================================================
ostream& operator <<(ostream &os, const BioSequence& seq) {
   seq.print(os);
   return os;
}
//==========================================================
istream& operator >>(istream &is, BioSequence& seq) {
   seq.read(is);
   return is;
}
/** ====================================== */
BioSubSequence::BioSubSequence() :
   BioSequence() {
   m_iFrom = m_iTo = 0;
}
/** ====================================== */
BioSubSequence::BioSubSequence(const BioSequence& other) :
   BioSequence(other) {
}
/** ====================================== */
BioSubSequence::BioSubSequence(const BioSubSequence& other) :
   BioSequence((BioSequence) other), m_iFrom(other.m_iFrom), m_iTo(other.m_iTo), m_sComment(other.m_sComment) {
}
/** ====================================== */
BioSubSequence::BioSubSequence(const string &sSeq, const string &sName, long iFrom, long iTo) :
   BioSequence(sSeq, sName) {
   fragments.clear();
   BioFragment newFragment(sName, iFrom, iTo);
   fragments.assign(1, newFragment);
   m_iFrom = iFrom;
   m_iTo = iTo;
}
/** ====================================== */
BioSubSequence::BioSubSequence(const char *pcSeq, const char *pcName, long iFrom, long iTo) :
   BioSequence(pcSeq, pcName) {
   fragments.clear();
   BioFragment newFragment(pcName, iFrom, iTo);
   fragments.assign(1, newFragment);
   m_iFrom = iFrom;
   m_iTo = iTo;
}
/** ====================================== */
long BioSubSequence::getFrom() const {
   if (fragments.size()) {
      return fragments.front().first;
   }
   return m_iFrom;
}
/** ====================================== */
long BioSubSequence::getTo() const {
   if (fragments.size()) {
      return fragments.back().second;
   }
   return m_iTo;
}
/** ====================================== */
const string& BioSubSequence::getComment() const {
   return m_sComment;
}
/** ====================================== */
void BioSubSequence::setFrom(long iFrom) {
   m_iFrom = iFrom;
   if (fragments.size()) {
      fragments.front().first = iFrom;
   } else {
      BioFragment newFragment(getName(), iFrom, iFrom);
      fragments.push_back(newFragment);
   }
}
/** ====================================== */
void BioSubSequence::setTo(long iTo) {
   m_iTo = iTo;
   if (fragments.size()) {
      fragments.back().second = iTo;
   } else {
      BioFragment newFragment(getName(), iTo, iTo);
      fragments.push_back(newFragment);
   }
}
/** ====================================== */
void BioSubSequence::addComment(const string& sComment) {
   m_sComment.append(sComment);
}
/** ====================================== */
void BioSubSequence::setComment(const string& sComment) {
   m_sComment = sComment;
}
/** ====================================== */
void BioSubSequence::setComment(const char *pcComment) {
   m_sComment = pcComment;
}
/** ====================================== */
void BioSubSequence::printHead(ostream& os) const {
#ifdef FASTA
   printFASTAHead(os);
#else

   printIGHead(os);
#endif

}
/** ====================================== */
void BioSubSequence::printHead(FILE* FOUT) const {
#ifdef FASTA
   printFASTAHead(FOUT);
#else

   printIGHead(FOUT);
#endif

}
/** ====================================== */
void BioSubSequence::printIGHead(ostream& os) const {
   if (getComment().size() > 0) {
      os << "; " << getComment() << "\n";
   }
   if (getFrom() * getTo() != 0) {
      os << "; FRAGMENT " << getFrom() << " -> " << getTo() << "\n" << m_sName << "\n";
      ;
   } else {
      os << ";\n" << m_sName << "\n";
      ;
   }
}
/** ====================================== */
void BioSubSequence::printIGHead(FILE* FOUT) const {
   if (getComment().size() > 0) {
      fprintf(FOUT, "; %s\n", getComment().c_str());
   }
   if (getFrom() * getTo() != 0) {
      fprintf(FOUT, "; FRAGMENT %ld -> %ld\n%s\n", getFrom(), getTo(), m_sName.c_str());
   } else {
      fprintf(FOUT, ";\n%s\n", m_sName.c_str());
   }
}
/** ====================================== */
void BioSubSequence::printFASTAHead(ostream& os) const {
   os << ">" << m_sName;
   if (getFrom() * getTo() != 0) {
      os << " FRAGMENT " << getFrom() << " -> " << getTo();
   }
   if (getComment().size() > 0) {
      os << " " << getComment();
   }
   os << "\n";
}
/** ====================================== */
void BioSubSequence::printFASTAHead(FILE* FOUT) const {
   if (getFrom() * getTo() != 0) {
      fprintf(FOUT, ">%s FRAGMENT %ld -> %ld", m_sName.c_str(), getFrom(), getTo());
   } else {
      fprintf(FOUT, ">%s", m_sName.c_str());
   }
   if (getComment().size() > 0) {
      fprintf(FOUT, " %s", getComment().c_str());
   }
   fprintf(FOUT, "\n");
}
/** ====================================== */
BioSubSequence& BioSubSequence::assign(const BioSequence& sSeq) {
   BioSequence::assign(sSeq);
   return *this;
}
/** ====================================== */
BioSubSequence& BioSubSequence::assign(const BioSubSequence& sSeq) {
   setFrom(sSeq.getFrom());
   setTo(sSeq.getTo());
   setComment(sSeq.getComment());
   BioSequence::assign(sSeq);
   return *this;
}
/** No descriptions */
SEQTYPE BioSequence::guessType() const {
   unsigned long piStat[BASERANGE];
   int iSize = 0;
   string::const_iterator sptr;
   for (int i = 0; i < BASERANGE; i++) {
      piStat[i] = 0;
   }
   for (sptr = begin(); sptr != end(); sptr++) {
      if (*sptr < BASERANGE) {
         if (isalpha(*sptr)) {
            piStat[::toupper(*sptr)]++;
            iSize++;
         }
      }
   }
   if (piStat['A'] + piStat['T'] + piStat['C'] + piStat['G'] > 0) {
      if ((double) (piStat['A'] + piStat['T'] + piStat['C'] + piStat['G']) / (iSize - piStat['N'] - piStat['X']) > 0.75) {
         return NA;
      }
   } else {
      if (((double) (piStat['N'])) / (iSize) > 0.75) {
         return NA;
      }
   }
   return AA;
}
/** No descriptions */
int BioSequence::readHead(FILE* is) {
   char c;
   char sLine[MAX_LINELENGTH];
   char sName[MAX_LINELENGTH];
   if (is == NULL) {
      return 0;
   }
   while ((c = fgetc(is)) == '\n')
      ;

   if (((c = ungetc(c, is)) == ';' || c == '>') && fgets(sLine, MAX_LINELENGTH, is) != 0) {
      switch (sLine[0]) {
      case '>':
         sscanf(sLine + 1, "%s", sName);
         break;
      case ';':
         while (fgets(sLine, MAX_LINELENGTH, is) != 0 && sLine[0] == ';') {
            sLine[0] = '\0';
         }
         sscanf(sLine, "%s", sName);
         break;
      default:
         throw *(new baddata_exception("BioSequence::readHead(FILE*) : NO '>' or ';' FOUND.\n"));
         return 0;
      }
      setName(sName);
      return 1;
   } else {
      return 0;
   }
}
/** No descriptions */
int BioSequence::readHead(istream& is) {
   string sLine;
   string sName;
   istringstream iss;
   char c;
   while (is.peek() == '\n') {
      getline(is, sLine);
   }
   if ((is.peek() == ';' || is.peek() == '>') && !(getline(is, sLine).eof())) {
      switch (sLine[0]) {
      case '>':
         iss.clear();
         iss.str(sLine);
         iss >> c >> sName;
         break;
      case ';':
         while (!(getline(is, sLine).eof()) && sLine[0] == ';')
            ;
         iss.clear();
         iss.str(sLine);
         iss >> sName;
         break;
      default:
         throw *(new baddata_exception("BioSequence::readHead(istream*) : NO '>' or ';' FOUND.\n"));
         return 0;
      }
      setName(sName);
      return 1;
   } else {
      return 0;
   }
}
/** No descriptions */
int BioSubSequence::readHead(FILE* is) {
   char c;
   char *cptr;
   long int iF, iT;
   char sLine[MAX_LINELENGTH];
   char sName[MAX_LINELENGTH];
   m_iFrom = 0;
   m_iTo = 0;

   if (is == NULL) {
      return 0;
   }
   while ((c = fgetc(is)) == '\n')
      ;

   if (((c = ungetc(c, is)) == ';' || c == '>') && fgets(sLine, MAX_LINELENGTH, is) != 0) {
      switch (sLine[0]) {
      case '>':
         sscanf(sLine + 1, "%s", sName);
         if ((cptr = strstr(sLine, FRAGMENT)) != 0) {
            if (sscanf(cptr + strlen(FRAGMENT), "%ld -> %ld", &iF, &iT) == 2) {
               m_iFrom = iF;
               m_iTo = iT;
            }
         }
         break;
      case ';':
         do {
            if ((cptr = strstr(sLine, FRAGMENT)) != 0) {
               if (sscanf(cptr + strlen(FRAGMENT), "%ld -> %ld", &iF, &iT) == 2) {
                  m_iFrom = iF;
                  m_iTo = iT;
               }
            }
            sLine[0] = '\0';
         } while (fgets(sLine, MAX_LINELENGTH, is) != 0 && sLine[0] == ';');
         sscanf(sLine, "%s", sName);
         break;
      default:
         throw *(new baddata_exception("BioSubSequence::readHead(FILE*) : NO '>' or ';' FOUND.\n"));
         return 0;
      }
      setName(sName);
      return 1;
   } else {
      return 0;
   }
}
/** No descriptions */
int BioSubSequence::readHead(istream& is) {
   char *cptr;
   long int iF, iT;
   char sLine[MAX_LINELENGTH];
   char sName[MAX_LINELENGTH];
   m_iFrom = 0;
   m_iTo = 0;
   while (is.peek() == '\n') {
      is.getline(sLine, MAX_LINELENGTH);
   }
   if ((is.peek() == ';' || is.peek() == '>') && !is.getline(sLine, MAX_LINELENGTH).eof()) {
      switch (sLine[0]) {
      case '>':
         sscanf(sLine + 1, "%s", sName);
         if ((cptr = strstr(sLine, FRAGMENT)) != 0) {
            if (sscanf(cptr + strlen(FRAGMENT), "%ld -> %ld", &iF, &iT) == 2) {
               m_iFrom = iF;
               m_iTo = iT;
            }
         }
         break;
      case ';':
         do {
            if ((cptr = strstr(sLine, FRAGMENT)) != 0) {
               if (sscanf(cptr + strlen(FRAGMENT), "%ld -> %ld", &iF, &iT) == 2) {
                  m_iFrom = iF;
                  m_iTo = iT;
               }
            }
            sLine[0] = '\0';
         } while (!is.getline(sLine, MAX_LINELENGTH).eof() && sLine[0] == ';');
         sscanf(sLine, "%s", sName);
         break;
      default:
         throw *(new baddata_exception("BioSubSequence::readHead(istream) : NO '>' or ';' FOUND.\n"));
         return 0;
      }
      setName(sName);
      return 1;
   } else {
      return 0;
   }
}
/** No descriptions */
BioSequence& BioSubSequence::trunc_left(unsigned int n) {
   if (m_iFrom * m_iTo != 0) {
      if (m_iFrom < m_iTo) {
         m_iFrom += effective_length(substr(0, n));
      } else {
         m_iFrom -= effective_length(substr(0, n));
      }
   }
   erase(0, n);
   return *this;
}
/** No descriptions */
BioSequence& BioSubSequence::trunc_right(unsigned int n) {
   if (m_iFrom * m_iTo != 0) {
      if (m_iFrom < m_iTo) {
         m_iTo -= effective_length(substr(size() - n));
      } else {
         m_iTo += effective_length(substr(size() - n));
      }
   }
   erase(size() - n);
   return *this;
}
/** No descriptions */
unsigned int BioSequence::effective_length(const string& s) {
   string::const_iterator sptr;
   unsigned int iLen = 0;
   for (sptr = s.begin(); sptr != s.end(); sptr++) {
      if (isalpha(*sptr)) {
         iLen++;
      }
   }
   return iLen;
}
/** No descriptions */
unsigned int BioSequence::effective_length() {
   return effective_length(*this);
}
/** No descriptions */
int BioSequence::baseCount2position(int n) {
   return baseCount2position(*this, n);
}
/** No descriptions */
int BioSequence::baseCount2position(const string& s, int n) {
   int p = 0;
   for (string::const_iterator sptr = s.begin(); sptr != s.end() && n > 0; sptr++) {
      if (isalpha(*sptr)) {
         n--;
      }
      p++;
   }
   return p;
}
size_t BioSequence::baseCount(size_t pos, int n) const {
   return baseCount(*this, pos, n);
}
size_t BioSequence::baseCount(const string& s, size_t pos, int n) {
   if (pos < 0 || pos >= s.size()) {
      return 0;
   }
   int mod = n < 0 ? 1 : -1;
   size_t start = pos + n;
   if (start < 0) {
      start = 0;
   }
   if (start >= s.size()) {
      start = s.size() - 1;
   }
   size_t cnt = (isalpha(s[pos] || s[pos] == '*')) ? 1 : 0;
   for (size_t i = start; i != pos; i += mod) {
      if (isalpha(s[i] || s[i] == '*')) {
         cnt++;
      }
   }
   return cnt;
}
/** */
BioSequence BioSequence::subseq(size_t pos, size_t n) const {
   return *(new BioSequence(subseq(*this, pos, n), getName()));
}
/** */
BioSubSequence BioSubSequence::subseq(size_t pos, size_t n, bool literal) const {
   int sign = getFrom() <= getTo() ? +1 : -1;
   if (!literal) {
      return *(new BioSubSequence(BioSequence::subseq(pos, n), getName(), getFrom() + sign * pos, getFrom() + sign * (pos + n)));
   }
   return *(new BioSubSequence(string::substr(pos, n), getName(), getFrom() + sign * baseCount(0, pos), getFrom() + sign * baseCount(0, pos + n)));
}
/** */
string BioSequence::subseq(const string& s, size_t pos, size_t n) {
   int beg = 0;
   int len = 0;
   while (pos + n > 0) {
      if (pos > 0) {
         beg++;
         if (isalpha(s[beg]) || s[beg] == '*') {
            pos--;
         }
      } else {
         len++;
         if (isalpha(s[beg + len]) || s[beg + len] == '*') {
            n--;
         }
      }
   }
   return s.substr(beg, len);
}
/** Read property of SEQTYPE seqtype. */
SEQTYPE BioSequence::getseqtype() {
   if (seqtype == SEQTYPE_UNKNOWN) {
      seqtype = guessType();
   }
   return seqtype;
}
string& BioSequence::shrink(string& s, char c) {
   if (s.size() < 2) {
      return s;
   }
   char lastChar = s[s.size() - 1];
   for (int i = s.size() - 2; i >= 0; i--) {
      if (s[i] == lastChar && s[i] == c) {
         s.erase(i, 1);
      } else {
         lastChar = s[i];
      }
   }
   return s;
}
/*
 * Modifies coordinates relative to the source sequence.
 * [start, end] - coordinates relative to this subsequence.
 * **/
void BioSubSequence::adjust_position(int start, int end) {
   if (getFrom() * getTo() == 0) {
      setFrom(start);
      setTo(end);
      return;
   }
   if (getFrom() <= getTo()) {
      setFrom(getFrom() + (start - 1));
      setTo(getFrom() + (end - start));
   } else {
      setFrom(getFrom() - (start - 1));
      setTo(getFrom() - (end - start));
   }
}
list<BioFragment>& BioSubSequence::getFragments() {
   return fragments;
}
BioSubSequence& BioSubSequence::append(BioSubSequence& other) {
   BioSequence::append((BioSequence) other);
   if (other.getFrom() * other.getTo() == 0) {
      return *this;
   }
   if (getFrom() * getTo() == 0) {
      getFragments().assign(other.getFragments().begin(), other.getFragments().end());
   } else {
      getFragments().insert(getFragments().end(), other.getFragments().begin(), other.getFragments().end());
   }
   return *this;
}
/** ====================================== */
void BioSubSequence::clear() {
   BioSequence::clear();
   setFrom(0);
   setTo(0);
   setComment("");
   getFragments().clear();
}

RatedSequence::RatedSequence() :
   score(0) {
}
RatedSequence::RatedSequence(const BioSequence& other) :
   BioSequence(other), score(0) {
}
RatedSequence::RatedSequence(const RatedSequence& other) :
   BioSequence(other), score(other.score) {
}
RatedSequence& RatedSequence::assign(const RatedSequence& other) {
   BioSequence::assign((BioSequence) other);
   score = other.score;
   return *this;
}
void RatedSequence::setScore(double score) {
   this->score = score;
}
double RatedSequence::getScore() const {
   return score;
}
void RatedSequence::clear() {
   BioSequence::clear();
   score = 0;
}
/** ====================================== */
void RatedSequence::printHead(ostream& os) const {
#ifdef FASTA
   printFASTAHead(os);
#else

   printIGHead(os);
#endif

}
/** ====================================== */
void RatedSequence::printHead(FILE* FOUT) const {
#ifdef FASTA
   printFASTAHead(FOUT);
#else

   printIGHead(FOUT);
#endif

}
/** ====================================== */
void RatedSequence::printIGHead(ostream& os) const {
   if (score) {
      os << "; SCORE= " << getScore() << endl;
   }
   BioSequence::printIGHead(os);
}
/** ====================================== */
void RatedSequence::printIGHead(FILE* FOUT) const {
   if (score) {
      fprintf(FOUT, "; SCORE= %f\n", getScore());
   }
   BioSequence::printIGHead(FOUT);
}
/** ====================================== */
void RatedSequence::printFASTAHead(ostream& os) const {
   os << ">" << m_sName;
   if (score) {
      os << " SCORE= " << getScore();
   }
   os << endl;
}
/** ====================================== */
void RatedSequence::printFASTAHead(FILE* FOUT) const {
   if (score) {
      fprintf(FOUT, ">%s SCORE= %f\n", m_sName.c_str(), getScore());
   } else {
      fprintf(FOUT, ">%s\n", m_sName.c_str());
   }
}
